////////////////////////////////////////////////////////////////////////
// Class:       Multiplet
// Module Type: filter
// File:        Multiplet_module.cc
//
// This is a slicer designed to keep only very small slices, with finding SN
// nu hits in mind.  It doesn't try to do 3D matching, just simply looks for
// things that are near in plane number and time.  Single hits are probably 
// best rejected: noise is predominantly singletons, even though so are almost
// half of your SN nu hits :(
//
// Generated at Tue Sep  9 04:40:50 2014 by Alec Habig using artmod
// from cetpkgsupport v1_07_00.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitSet.h"

#include <memory>
#include <list>
#include <algorithm>

//a small function to get distance for unsigned values
template<class T>
unsigned distance(const T& t1, const T& t2){
    return (t2>t1)?(t2-t1):(t1-t2);
}

template <class T>
unsigned delta(const novaddt::DAQHit& h1, const novaddt::DAQHit& h2){
    return distance(static_cast<T>(h1).val, static_cast<T>(h2).val);
}


namespace novaddt {
  class Multiplet;
}

class novaddt::Multiplet : public art::EDFilter {
public:
  explicit Multiplet(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  Multiplet(Multiplet const &) = delete;
  Multiplet(Multiplet &&) = delete;
  Multiplet & operator = (Multiplet const &) = delete;
  Multiplet & operator = (Multiplet &&) = delete;

  // Required functions.
  bool filter(art::Event & event) override;

  // Selected optional functions.
  void beginJob() override;
  void endJob() override;

  // Check if hit belongs to group
  bool isHitCloseToSlice(const DAQHit& h, const HitList& frag){
    for(const auto& h2: frag){
        if(isHitCloseToHit(h,h2))return true;
    }
    return false;
  }

  bool isHitCloseToHit( const DAQHit& h1, const DAQHit& h2){
    if (delta<Plane>(h1,h2)>fMaxPlaneSep) return false;
    return (delta<View>(h1,h2)!=0)||(delta<Cell>(h1,h2)<=fMaxCellSep);
  }

private:
  template<class HitCollection>
  void findslices(const HitCollection& inhits, std::unique_ptr< std::vector< novaddt::HitList > >& product);


private:
  // Declare member data here.
  art::InputTag   fInputTag;
  unsigned        fMaxPlaneSep;
  unsigned        fMaxCellSep;
  unsigned        fMaxTDCSep;
  unsigned        fMinSliceSize;
  unsigned        fMaxSliceSize;
  bool            fVerbose;

  // Counters
  int _nEvents        = 0;
  int _nHitsIn        = 0;
  int _nHitsOut       = 0;
  int _nSlicesOut     = 0;
};


novaddt::Multiplet::Multiplet(fhicl::ParameterSet const & p):
fInputTag       (p.get< std::string >("InputTag"   )),
fMaxPlaneSep    (p.get< unsigned    >("slicer.MaxPlaneSep"  )),
fMaxCellSep     (p.get< unsigned    >("slicer.MaxCellSep"  )),
fMaxTDCSep      (p.get< unsigned    >("slicer.MaxTDCSep"    )),
fMinSliceSize   (p.get< unsigned    >("slicer.MinSliceSize" )),
fMaxSliceSize   (p.get< unsigned    >("slicer.MaxSliceSize" )),
fVerbose        (p.get< bool        >("Verbose"      ))
// Initialize member data
{
    if (fVerbose){
      std::cout << "--- novaddt::Multiplet begin" << std::endl;
      std::cout << "\t Input hitlists tag:      " << fInputTag     << std::endl;

      std::cout << "\t Max Z separation:        " << fMaxPlaneSep   << std::endl;
      std::cout << "\t Max XY separation:       " << fMaxCellSep   << std::endl;
      std::cout << "\t Max time separation:     " << fMaxTDCSep     << std::endl;
      std::cout << "\t Min slice size:          " << fMinSliceSize  << std::endl;
      std::cout << "\t Max slice size:          " << fMaxSliceSize  << std::endl;
      std::cout << "\t Verbose:                 " << fVerbose       << std::endl;
    }
    
    // Call appropriate Produces<>() functions
    produces<std::vector<HitList>>();
}
//////////////////////////////////////////////////////////////////////

bool novaddt::Multiplet::filter(art::Event & event)
{
    _nEvents++;
    
    // Hitlist from event (input)
    auto hitsHandle=event.getValidHandle<novaddt::HitList>(fInputTag);

    novaddt::HitSet allHits(*hitsHandle);

    _nHitsIn+=allHits.size();

    //make a vector of hitlists out of hits after slicing
    std::unique_ptr< std::vector<HitList> >product     (new std::vector<HitList>);

    // Find slices
    findslices(allHits,product);
    
    _nSlicesOut+=product->size();
    for (auto slice: *product) _nHitsOut+=slice.size();

    // Add filtered hitlists to event
    event.put(std::move(product     ));
    return true;
}
void novaddt::Multiplet::beginJob()
{
  // Implementation of optional member function here.
}
//////////////////////////////////////////////////////////////////////
void novaddt::Multiplet::endJob()
{ 
    if (!fVerbose) return;
    std::cout << "--- novaddt::Multiplet end " << std::endl;
    std::cout << "\t# of events:        " << _nEvents   << std::endl;
    std::cout << "\t# of hits input:    " << _nHitsIn   << std::endl;
    std::cout << "\t# of hits output:   " << _nHitsOut  << std::endl;
    std::cout << "\t# of slices output: " << _nSlicesOut  << std::endl;
}

std::ostream& operator << (std::ostream& s, novaddt::DAQHit& hit){
    s<<"adc="<<hit.ADC().val
     <<" p="<<hit.Plane().val
     <<" c="<<hit.Cell().val<<"("<<hit.View().val<<")"
     <<" t="<<hit.TDC().val;
    return s;
}


 
//////////////////////////////////////////////////////////////////////
template<class HitCollection>
void novaddt::Multiplet::findslices(const HitCollection & inhits,
  std::unique_ptr<std::vector<HitList> > & product)
{
  if (!inhits.empty()) {

    // copy input hitlist to a list that we can start slicing up
    std::list<DAQHit> hits(inhits.begin(),inhits.end());

    if (fVerbose) std::cout << "findslices got " << inhits.size()
      << " hits, working with " << hits.size()<<std::endl;
    // Loop over all hits
    while (!hits.empty()) {
      // this will be the first hit in the slice, put it there
      HitList frag;
      frag.emplace_back(hits.front());
      // remove it from the front of the list, since it's now in a slice
      hits.pop_front();
      // point to the first hit in the slice
      auto lastInSlice = frag.back();
      auto testhit = hits.begin();
      while (testhit != hits.end()) {
        // since hits are time ordered, march along until we get out of time,
        // with the last hit already in the slice, adding hits that are 
        // also in space with the last hit already in the slice
        if(delta<TDC>(*testhit,lastInSlice) > fMaxTDCSep) break;
        //we're out of time, this fragment ends

       if(isHitCloseToSlice(*testhit,frag))
       {
               // add this hit to slice
                frag.emplace_back(*testhit);
                lastInSlice = frag.back();
                // remove it from list of potential hits, since a hit can only 
                // belong to one slice (after moving to next hit)
                auto erasehit = testhit;
                testhit++;
                hits.erase(erasehit);
         }
        else testhit++;
        
      } // end loop over rest of hits

      // we've finished building a slice, save it to product
      if (fVerbose) std::cout << "Sliced: " << frag.size() << " hits from " 
                << frag.front().TDC().val << " to "
                << frag.back().TDC().val <<std::endl;
      if ((frag.size() <= fMaxSliceSize)
       && (frag.size() >= fMinSliceSize)){
        product->emplace_back(frag);
      }
    } // loop over full hitlist

  }
  // Last slices seem to be noise slices.  
  // Save the slices that don't meet the cuts in here
  // TODO: right now we're just dropping "noise" hits, more efficient.  
  // Make an empty slice just in case
  //HitList noise;
  //product->emplace_back(noise);
}
//////////////////////////////////////////////////////////////////////
DEFINE_ART_MODULE(novaddt::Multiplet)
