#ifndef BUFFERMANAGER__H
#define BUFFERMANAGER__H
#include <boost/interprocess/managed_shared_memory.hpp>
#include <boost/interprocess/sync/interprocess_upgradable_mutex.hpp>
#include <boost/interprocess/sync/sharable_lock.hpp>
#include <boost/interprocess/sync/upgradable_lock.hpp>

namespace novaddt{
namespace calib{

        using mutexType=boost::interprocess::interprocess_upgradable_mutex;
        using sharableLock  =boost::interprocess::sharable_lock<mutexType>;
        using upgradableLock=boost::interprocess::upgradable_lock<mutexType>;
        using scopedLock    =boost::interprocess::scoped_lock<mutexType>;

template <class bufferType, uint16_t Nbuffers>
class BufferContainer{
   using indexType=uint16_t;
   public:
       //manipulate 
       void SwapBuffers(){
            //go to next writing position
            fIndex =Index(fIndex+1);
            WriteBuffer().reset();
       }
             bufferType& WriteBuffer(){return fBuffersArray[fIndex];}
       const bufferType& ReadBuffer(int i=0) const {return fBuffersArray[ReadIndex(i)];}
   protected:
        indexType Index(int idx) const {return indexType(idx%Nbuffers);}
        indexType ReadIndex(int idx) const{
            idx=1+idx%(Nbuffers-1);
            return Index(fIndex+Nbuffers-idx);
        }
   protected:
        //data container
        bufferType  fBuffersArray[Nbuffers];  //array of buffers in the shared memory [NBuffers]
        indexType   fIndex;        //point to buffer which we write
};

template <class bufferType, uint16_t Nbuffers>
class BufferManager{
    public:
        //default c-tor
        BufferManager(size_t CounterMax = 20 ):
                fCounter(0),
                fCounterMax(CounterMax),
                fBufferContainer(),
                fMutex()
                {}
        void SetCounterMax(size_t cmax){fCounterMax=cmax;}
        //obtain the sharable_lock this must be done only AFTER addCounter
        sharableLock getLock(){return sharableLock(fMutex);}
        
        //interface to the counter
        size_t getCounter() {return fCounter;}
        size_t getCounterMax() {return fCounterMax;}
        bool isCounterFull(){return fCounter>=fCounterMax;}
        //increment counter and try to swap Buffers in case of overflow
        void addCounter(size_t n=1){if(isCounterFull())tryToSwap(); fCounter+=n;}
        //interface to the buffers container
             bufferType& getWBuffer(){return fBufferContainer.WriteBuffer();}
       const bufferType& getRBuffer(int i=0) const {return fBufferContainer.ReadBuffer(i);}

    protected:
        void tryToSwap(){
           ///make sure this is the only process trying to swap: obtain upgradable_lock
           auto upg_lock =  upgradableLock(fMutex);
           if(!isCounterFull())return; //somebody has already taken care of swapping
           //if not - we have to deal with it. Lock exclusively, so other processes don't read or write
           auto exc_lock = scopedLock(std::move(upg_lock));
           fBufferContainer.SwapBuffers();
           fCounter=0;
        }
    protected:
        size_t fCounter = 0;
        size_t fCounterMax = 100;
        BufferContainer<bufferType,Nbuffers> fBufferContainer;
        mutexType fMutex;
};

}
}
#endif
