#ifndef CHANNELMAP_SERVICE__CC
#define CHANNELMAP_SERVICE__CC
////////////////////////////////////////////////////////////////////////
/// \file ChannelMap_service.cc
///  
/// \brief A service to manipulate channel map, containing activity
///
/// \author Andrey Sheshukov
////////////////////////////////////////////////////////////////////////

#include "DDTBaseDataProducts/HitList.h"

#include "OnlineCalibration/ChannelMapService.h"
#include <boost/algorithm/string/replace.hpp>
#include <boost/lexical_cast.hpp>

namespace novaddt{
MapSource::MapSource(fhicl::ParameterSet const& p):
    type(p.get<std::string>("type")),
    name(p.get<std::string>("name")),
    count(p.get<int>("count"))
    {
        if(type!="file" && type!="product")
        mf::LogError("MapSource")<<"Wrong source type '"<<type<<"'";
    }
}


std::string MakeFileName(const std::string fmt, const art::Run& run){
    std::string result=fmt;    
    std::string RunS=boost::lexical_cast<std::string>(run.run());
    boost::replace_all(result,"%r",RunS);
    return result;
};

namespace novaddt{
ChannelMapService::ChannelMapService
    (fhicl::ParameterSet const& p, art::ActivityRegistry& reg):
    fMapSource(p.get<fhicl::ParameterSet>("source")),
    fMapSave(p.get<std::string>("save.filename")),
    fSegName(p.get<std::string>("shmem.segment")),
    fMapName(p.get<std::string>("shmem.object")),
    fSharedMapContainer(fSegName.data(),fMapName.data())
{
    mf::LogInfo("ChannelMapService")<<"\t config:"
    <<"\n\tmap source:"
    <<"\n\t\ttype:"<<fMapSource.type
    <<"\n\t\tname:"<<fMapSource.name
    <<"\n\t\tcount:"<<fMapSource.count
    <<"\nshmem.segment="<<fSegName<<";shmem.object="<<fMapName;
 
    fSharedMapContainer.getData().SetCounterMax(fMapSource.count);
    if(fMapSource.type=="file")
        readMapFromFile(fMapSource.name);

    //register callbacks: before starting run
    reg.sPreBeginRun.watch(&novaddt::ChannelMapService::preBeginRun, *this);
     
    //register callbacks: after processing an event
    if(fMapSource.type=="product")
        reg.sPostProcessEvent.watch(&novaddt::ChannelMapService::postProcessEvent, *this);
}

void ChannelMapService::preBeginRun(const art::Run& run){
    
    saveMapToFile(MakeFileName(fMapSave,run));
}

void ChannelMapService::postProcessEvent(const art::Event& event)
{
   fillMapFromEvent(event,fMapSource.name);
}

void ChannelMapService::readMapFromFile(const std::string& fname)
{
        mf::LogInfo("ChannelMapService")<<"read map from file \""<<fname<<"\"";

        std::ifstream s(fname.data(),std::ifstream::binary);
        if(!s.good()){
            mf::LogError("ChannelMapService")<<" error opening file \""<<fname<<"\"";
            return;
        }
        
        {
            auto lock= GetLock();
            std::cout<<"Let's start!"<<std::endl;
            auto &map = GetWMap();
            std::cout<<"sizeof(map)="<<sizeof(map)
                    <<" ["<<sizeof(map)/map.Length()<<" per channel ]"<<std::endl;
            size_t length = sizeof(map);
            char buffer[length];
            //reinterpret_cast<char*>(&map);
            s.read(buffer,length);
            mf::LogInfo("ChannelMapService")<<"read successfully";
        }
}
void ChannelMapService::saveMapToFile(const std::string& fname)
{
        mf::LogInfo("ChannelMapService")<<"save map to file \""<<fname<<"\"";

        std::ofstream s(fname.data(),std::ofstream::binary);
        if(!s.good()){
            mf::LogError("ChannelMapService")<<" error opening file \""<<fname<<"\"";
            return;
        }
        
        auto lock= GetLock();
        auto map = GetRMap();
        std::cout<<"sizeof(map)="<<sizeof(map)
                <<" ["<<sizeof(map)/map.Length()<<" per channel ]"<<std::endl;;
        s.write(reinterpret_cast<char*>(&map),sizeof(map));
}

void ChannelMapService::fillMapFromEvent
       (const art::Event& event, const art::InputTag& product)
{
    auto hits=event.getValidHandle<novaddt::HitList>(product);
    mf::LogDebug("ChannelMapService")<<"fill map from \""<<product
        <<"\" with "<<hits->size()<<" hits"
        <<"\n Counter="<<getCounter();
    
    addCounter();
    auto lock = GetLock();
    auto &NoiseMap = GetWMap();
    //fill the hits map
    for(const auto &hit: *hits)
        NoiseMap[hit]+=1;
}


calib::sharableLock ChannelMapService::GetLock(){
    std::cout<<"+++requested lock+++"<<std::endl;
    return fSharedMapContainer.getData().getLock();
}

const calib::FDHitMap& ChannelMapService::GetRMap() const{
    return fSharedMapContainer.getData().getRBuffer();
}

calib::FDHitMap& ChannelMapService::GetWMap(){
    return fSharedMapContainer.getData().getWBuffer();
}


DEFINE_ART_SERVICE(ChannelMapService);
}
#endif
